#!/usr/bin/env python3
"""Tacomail CLI - Command-line interface for Tacomail disposable email service."""

from typing import Optional, Any, Callable
from enum import Enum
import re
import json
from datetime import datetime

import typer
from rich.console import Console
from rich.table import Table
from rich.panel import Panel

from tacomail import (
    TacomailClient,
    AsyncTacomailClient,
    Email,
)


class OutputFormat(str, Enum):
    """Output format options for CLI commands."""
    RICH = "rich"
    PLAIN = "plain"
    JSON = "json"


app = typer.Typer(
    name="tacomail",
    help="Tacomail CLI - Disposable email service command-line interface",
    no_args_is_help=True,
)

console = Console()

# Global options
use_async = False
output_format: OutputFormat = OutputFormat.RICH


def output_data(data: dict[str, Any] | list[Any], rich_callback: Callable[[], None]) -> None:
    """Output data in the configured format.

    Args:
        data: The data to output (dict for single items, list for multiple items)
        rich_callback: A callback function to render rich output (only called for RICH format)
    """
    if output_format == OutputFormat.JSON:
        print(json.dumps(data, default=str))
    elif output_format == OutputFormat.PLAIN:
        if isinstance(data, list):
            for item in data:
                if isinstance(item, dict):
                    # For list of dicts, output tab-separated values
                    print("\t".join(str(v) for v in item.values()))
                else:
                    print(item)
        else:
            for key, value in data.items():
                print(f"{key}={value}")
    else:
        rich_callback()


def get_client():
    """Get the appropriate client based on global async setting."""
    if use_async:
        return AsyncTacomailClient()
    return TacomailClient()


# Email commands
@app.command()
def create(
    domain: Optional[str] = typer.Option(
        None, "--domain", "-d", help="Specific domain to use (otherwise random)"
    ),
    username: Optional[str] = typer.Option(
        None, "--username", "-u", help="Specific username to use (otherwise random)"
    ),
) -> None:
    """Create a random email address.

    Generates a random email address using available domains.
    You can optionally specify a domain or username.
    """
    client = get_client()

    try:
        if username and domain:
            email_address = f"{username}@{domain}"
        elif domain:
            username = client.get_random_username()
            email_address = f"{username}@{domain}"
        else:
            email_address = client.get_random_address()

        def rich_output():
            console.print(Panel(
                f"[bold green]Generated Email:[/bold green]\n{email_address}",
                title="✨ Success",
                border_style="green"
            ))

        output_data({"email": email_address}, rich_output)
    except Exception as e:
        console.print(f"[red]Error creating email:[/red] {e}")
        raise typer.Exit(1)


@app.command("list-domains")
def list_domains() -> None:
    """List all available domains for email addresses.

    Shows all domains that can be used to create email addresses.
    """
    client = get_client()

    try:
        domains = client.get_domains()

        def rich_output():
            console.print(f"\n[bold]Available Domains:[/bold] ({len(domains)})")
            for domain in domains:
                console.print(f"  • {domain}")

        output_data(domains, rich_output)
    except Exception as e:
        console.print(f"[red]Error fetching domains:[/red] {e}")
        raise typer.Exit(1)


# Session commands
@app.command("create-with-session")
def create_with_session(
    domain: Optional[str] = typer.Option(
        None, "--domain", "-d", help="Specific domain to use (otherwise random)"
    ),
    username: Optional[str] = typer.Option(
        None, "--username", "-u", help="Specific username to use (otherwise random)"
    ),
) -> None:
    _create_with_session_impl(domain, username)


@app.command("new")
def new_email_with_session(
    domain: Optional[str] = typer.Option(
        None, "--domain", "-d", help="Specific domain to use (otherwise random)"
    ),
    username: Optional[str] = typer.Option(
        None, "--username", "-u", help="Specific username to use (otherwise random)"
    ),
) -> None:
    """Create a random email address and session in one command (alias for 'create-with-session').

    This is a short alias for the 'create-with-session' command.
    Use 'tacomail new' instead of 'tacomail create-with-session' for convenience.

    Example:
        tacomail new

    With options:
        tacomail new --domain tacomail.de
        tacomail new -u myuser -d tacomail.de
    """
    _create_with_session_impl(domain, username)


def _create_with_session_impl(
    domain: Optional[str],
    username: Optional[str],
) -> None:
    """Create a random email address and session in one command.

    This command combines the functionality of 'create' and 'create-session' into a single,
    convenient operation. It generates a random email address and automatically creates a session
    for it, eliminating the need to run two separate commands when you want to start receiving
    emails immediately.

    The command returns both the email address and session information, including the expiration
    time. This is the most efficient way to set up a temporary email inbox.

    Example:
        tacomail create-with-session

    With options:
        tacomail create-with-session --domain tacomail.de
        tacomail create-with-session --username myuser --domain tacomail.de

    This command works in both sync and async modes (use --async flag for async mode).
    """
    client = get_client()

    def display_results(email_address: str, session, expires_str: str) -> None:
        """Display the results using the configured output format."""
        data = {
            "email": email_address,
            "username": session.username,
            "domain": session.domain,
            "expires": expires_str,
        }

        def rich_output():
            console.print(Panel(
                f"[bold green]Email Address:[/bold green]\n{email_address}\n\n"
                f"[bold green]Session Created[/bold green]\n\n"
                f"[bold]Expires:[/bold] {expires_str}\n"
                f"[bold]Username:[/bold] {session.username}\n"
                f"[bold]Domain:[/bold] {session.domain}\n\n"
                f"[dim]You can now receive emails at this address![/dim]",
                title="✨ Email & Session Ready",
                border_style="green"
            ))

            console.print("\n[bold cyan]Next steps:[/bold cyan]")
            console.print("  • Monitor inbox: [green]tacomail list {}[/green]".format(email_address))
            console.print("  • Wait for email: [green]tacomail wait {}[/green]".format(email_address))

        output_data(data, rich_output)

    try:
        # Handle async mode
        if use_async:
            import asyncio

            async def async_operation():
                # Generate email address
                nonlocal username
                if username and domain:
                    email_address = f"{username}@{domain}"
                elif domain:
                    username = await client.get_random_username()
                    email_address = f"{username}@{domain}"
                else:
                    email_address = await client.get_random_address()

                # Parse email for session creation
                if "@" not in email_address:
                    console.print(f"[red]Invalid email address generated:[/red] {email_address}")
                    raise typer.Exit(1)

                user, dom = email_address.split("@", 1)

                # Create session
                session = await client.create_session(user, dom)

                # Format expiration time (timestamp is in milliseconds)
                expires_dt = datetime.fromtimestamp(session.expires / 1000)
                expires_str = expires_dt.strftime("%Y-%m-%d %H:%M:%S")

                display_results(email_address, session, expires_str)

            asyncio.run(async_operation())
        else:
            # Generate email address
            if username and domain:
                email_address = f"{username}@{domain}"
            elif domain:
                username = client.get_random_username()
                email_address = f"{username}@{domain}"
            else:
                email_address = client.get_random_address()

            # Parse email for session creation
            if "@" not in email_address:
                console.print(f"[red]Invalid email address generated:[/red] {email_address}")
                raise typer.Exit(1)

            user, dom = email_address.split("@", 1)

            # Create session
            session = client.create_session(user, dom)

            # Format expiration time (timestamp is in milliseconds)
            expires_dt = datetime.fromtimestamp(session.expires / 1000)
            expires_str = expires_dt.strftime("%Y-%m-%d %H:%M:%S")

            display_results(email_address, session, expires_str)

    except Exception as e:
        console.print(f"[red]Error creating email and session:[/red] {e}")
        raise typer.Exit(1)


@app.command()
def create_session(
    email: str = typer.Argument(..., help="Email address (e.g., user@domain.com)"),
) -> None:
    """Create a session for an email address.

    Creating a session is required to receive emails.
    Only incoming emails with an associated session are saved.
    """
    client = get_client()

    try:
        # Parse email address
        if "@" not in email:
            console.print(f"[red]Invalid email address:[/red] {email}")
            raise typer.Exit(1)

        username, domain = email.split("@", 1)

        session = client.create_session(username, domain)

        expires_dt = datetime.fromtimestamp(session.expires / 1000)
        expires_str = expires_dt.strftime("%Y-%m-%d %H:%M:%S")

        data = {
            "email": email,
            "username": session.username,
            "domain": session.domain,
            "expires": expires_str,
        }

        def rich_output():
            console.print(Panel(
                f"[bold green]Session Created[/bold green]\n\n"
                f"Email: {email}\n"
                f"Expires: {expires_str}\n"
                f"Username: {session.username}\n"
                f"Domain: {session.domain}",
                title="🔐 Session",
                border_style="green"
            ))

        output_data(data, rich_output)
    except Exception as e:
        console.print(f"[red]Error creating session:[/red] {e}")
        raise typer.Exit(1)


@app.command()
def delete_session(
    email: str = typer.Argument(..., help="Email address (e.g., user@domain.com)"),
) -> None:
    """Delete a session for an email address.

    This will cause incoming emails to be rejected.
    Already saved emails are not deleted.
    """
    client = get_client()

    try:
        if "@" not in email:
            console.print(f"[red]Invalid email address:[/red] {email}")
            raise typer.Exit(1)

        username, domain = email.split("@", 1)

        client.delete_session(username, domain)

        def rich_output():
            console.print(f"[green]✓ Session deleted for {email}[/green]")

        output_data({"deleted": email}, rich_output)
    except Exception as e:
        console.print(f"[red]Error deleting session:[/red] {e}")
        raise typer.Exit(1)


# Inbox commands
@app.command("list")
def list_inbox(
    email: str = typer.Argument(..., help="Email address to check"),
    limit: Optional[int] = typer.Option(
        None, "--limit", "-l", help="Maximum number of emails to show (max 10)"
    ),
) -> None:
    """List emails in the inbox.

    Shows recent emails for the specified address.
    Maximum of 10 emails can be retrieved.
    """
    client = get_client()

    try:
        emails = client.get_inbox(email, limit=limit)

        if not emails:
            if output_format != OutputFormat.RICH:
                # For non-rich formats, output empty list
                output_data([], lambda: None)
                return
            console.print(f"[yellow]No emails found for {email}[/yellow]")
            return

        # Prepare data for output
        emails_data = [
            {
                "id": email_obj.id,
                "from": email_obj.from_.address,
                "subject": email_obj.subject,
                "date": email_obj.date.strftime("%Y-%m-%d %H:%M"),
            }
            for email_obj in emails
        ]

        def rich_output():
            table = Table(title=f"Inbox for {email}")
            table.add_column("ID", style="cyan", no_wrap=True)
            table.add_column("From", style="green")
            table.add_column("Subject", style="white")
            table.add_column("Date", style="blue")

            for email_obj in emails:
                from_addr = f"{email_obj.from_.name} <{email_obj.from_.address}>"
                subject = email_obj.subject[:50] + "..." if len(email_obj.subject) > 50 else email_obj.subject
                date_str = email_obj.date.strftime("%Y-%m-%d %H:%M")

                table.add_row(email_obj.id, from_addr, subject, date_str)

            console.print(table)
            console.print(f"\n[dim]Showing {len(emails)} email(s)[/dim]")

        output_data(emails_data, rich_output)
    except Exception as e:
        console.print(f"[red]Error listing inbox:[/red] {e}")
        raise typer.Exit(1)


@app.command("get")
def get_email(
    email: str = typer.Argument(..., help="Email address"),
    mail_id: str = typer.Argument(..., help="Email ID to retrieve"),
) -> None:
    """Get a specific email by ID.

    Displays full email details including body and attachments.
    """
    client = get_client()

    try:
        email_obj = client.get_email(email, mail_id)

        # Create formatted output
        from_addr = f"{email_obj.from_.name} <{email_obj.from_.address}>"
        to_addr = f"{email_obj.to.name} <{email_obj.to.address}>"
        date_str = email_obj.date.strftime("%Y-%m-%d %H:%M:%S")

        data = {
            "id": email_obj.id,
            "from": email_obj.from_.address,
            "from_name": email_obj.from_.name,
            "to": email_obj.to.address,
            "to_name": email_obj.to.name,
            "subject": email_obj.subject,
            "date": date_str,
            "attachments": len(email_obj.attachments),
            "attachment_files": [att.fileName for att in email_obj.attachments],
            "body": email_obj.body.text or "",
        }

        def rich_output():
            content = f"""
[bold]From:[/bold] {from_addr}
[bold]To:[/bold] {to_addr}
[bold]Subject:[/bold] {email_obj.subject}
[bold]Date:[/bold] {date_str}
[bold]ID:[/bold] {email_obj.id}

[bold]Attachments:[/bold] {len(email_obj.attachments)} file(s)
"""

            if email_obj.attachments:
                for att in email_obj.attachments:
                    status = "✓" if att.present else "✗"
                    content += f"  {status} {att.fileName} (ID: {att.id})\n"

            content += "\n[bold]Body:[/bold]\n"
            if email_obj.body.text:
                content += f"\n{email_obj.body.text}\n"
            else:
                content += "[dim](No text body)[/dim]\n"

            console.print(Panel(content.strip(), title="📧 Email", border_style="blue"))

        output_data(data, rich_output)
    except Exception as e:
        console.print(f"[red]Error getting email:[/red] {e}")
        raise typer.Exit(1)


@app.command("delete")
def delete_email(
    email: str = typer.Argument(..., help="Email address"),
    mail_id: str = typer.Argument(..., help="Email ID to delete"),
) -> None:
    """Delete a specific email by ID.

    Permanently removes the email from the inbox.
    """
    client = get_client()

    try:
        client.delete_email(email, mail_id)

        def rich_output():
            console.print(f"[green]✓ Email {mail_id} deleted[/green]")

        output_data({"deleted": mail_id}, rich_output)
    except Exception as e:
        console.print(f"[red]Error deleting email:[/red] {e}")
        raise typer.Exit(1)


@app.command("clear")
def clear_inbox(
    email: str = typer.Argument(..., help="Email address"),
    confirm: bool = typer.Option(
        False, "--yes", "-y", help="Skip confirmation prompt"
    ),
) -> None:
    """Delete all emails from the inbox.

    Permanently removes all emails from the specified address.
    """
    client = get_client()

    try:
        if not confirm and output_format == OutputFormat.RICH:
            typer.confirm(f"Delete all emails from {email}?", abort=True)

        client.delete_inbox(email)

        def rich_output():
            console.print(f"[green]✓ Inbox cleared for {email}[/green]")

        output_data({"cleared": email}, rich_output)
    except Exception as e:
        console.print(f"[red]Error clearing inbox:[/red] {e}")
        raise typer.Exit(1)


# Wait commands
@app.command()
def wait(
    email: str = typer.Argument(..., help="Email address to monitor"),
    timeout: int = typer.Option(
        30, "--timeout", "-t", help="Maximum wait time in seconds"
    ),
    interval: int = typer.Option(
        2, "--interval", "-i", help="Check interval in seconds"
    ),
    filter_pattern: Optional[str] = typer.Option(
        None, "--filter", "-f", help="Filter by subject or sender (regex pattern)"
    ),
    print_body: bool = typer.Option(
        False, "--print-body", "-b", help="Also print email body"
    ),
) -> None:
    """Wait for a new email to arrive.

    Monitors the inbox and waits for a new email.
    Optionally filter by subject or sender using regex.
    """
    client = get_client()

    try:
        if output_format == OutputFormat.RICH:
            console.print(f"[dim]Waiting for email to {email}... (timeout: {timeout}s)[/dim]")

        if filter_pattern:
            # Create filter function
            regex = re.compile(filter_pattern, re.IGNORECASE)

            def filter_fn(email_obj: Email) -> bool:
                return bool(
                    regex.search(email_obj.subject) or
                    regex.search(email_obj.from_.address) or
                    regex.search(email_obj.from_.name)
                )

            email_obj = client.wait_for_email_filtered(
                email,
                filter_fn=filter_fn,
                timeout=timeout,
                interval=interval
            )
        else:
            email_obj = client.wait_for_email(
                email,
                timeout=timeout,
                interval=interval
            )

        if email_obj:
            date_str = email_obj.date.strftime("%Y-%m-%d %H:%M:%S")
            data = {
                "id": email_obj.id,
                "from": email_obj.from_.address,
                "from_name": email_obj.from_.name,
                "subject": email_obj.subject,
                "date": date_str,
            }
            if print_body:
                data["body"] = email_obj.body.text or ""

            def rich_output():
                console.print("\n[green]✓ Email received![/green]")
                console.print(f"  From: {email_obj.from_.name} <{email_obj.from_.address}>")
                console.print(f"  Subject: {email_obj.subject}")

                # Print email body if requested
                if print_body:
                    if email_obj.body.text:
                        console.print("\n[bold]Email Body:[/bold]")
                        console.print(f"{email_obj.body.text}")
                    else:
                        console.print("\n[dim]No text body available[/dim]")

            output_data(data, rich_output)
        else:
            def rich_output():
                console.print("\n[yellow]⏱ Timeout: No email received[/yellow]")

            output_data({"timeout": True}, rich_output)
            raise typer.Exit(1)
    except typer.Exit:
        raise
    except Exception as e:
        console.print(f"[red]Error waiting for email:[/red] {e}")
        raise typer.Exit(1)


@app.callback()
def main(
    async_mode: bool = typer.Option(
        False, "--async", help="Use async client for operations"
    ),
    verbose: bool = typer.Option(
        False, "--verbose", "-v", help="Enable verbose output"
    ),
    output: OutputFormat = typer.Option(
        OutputFormat.RICH, "--output", "-o", help="Output format (rich, plain, json)"
    ),
) -> None:
    """Tacomail CLI - Disposable email service command-line interface."""
    global use_async, output_format
    use_async = async_mode
    output_format = output

    if verbose:
        import logging
        logging.basicConfig(level=logging.DEBUG)


if __name__ == "__main__":
    app()



import pytest
from tacomail import TacomailClient, Email, Session
from email_sender import PostmarkEmailSender


@pytest.fixture
def client():
    with TacomailClient("https://tacomail.de") as client:
        yield client


def test_create_session(client: TacomailClient):
    username = client.get_random_username()
    domains = client.get_domains()
    domain = domains[0]

    session = client.create_session(username, domain)

    assert isinstance(session, Session)
    assert session.username == username
    assert session.domain == domain
    assert isinstance(session.expires, int)
    assert session.expires > 0


def test_delete_session(client: TacomailClient):
    username = client.get_random_username()
    domains = client.get_domains()
    domain = domains[0]

    # Create a session first
    client.create_session(username, domain)

    # Delete the session (should not raise an exception)
    client.delete_session(username, domain)


def test_get_contact_email(client):
    email = client.get_contact_email()
    assert isinstance(email, str)
    assert "@" in email


def test_get_random_username(client):
    username = client.get_random_username()
    assert isinstance(username, str)
    assert len(username) > 0


def test_get_domains(client):
    domains = client.get_domains()
    assert isinstance(domains, list)
    assert len(domains) > 0
    assert all(isinstance(domain, str) for domain in domains)
    assert all("." in domain for domain in domains)


def test_fetch_empty_inbox(client: TacomailClient):
    # Get a random email address using available domains
    domains = client.get_domains()
    username = client.get_random_username()
    email = f"{username}@{domains[0]}"
    # Fetch inbox for this random email
    emails = client.get_inbox(email)
    assert isinstance(emails, list)
    assert len(emails) == 0


@pytest.mark.flow
def test_email_sender_initialization():
    sender = PostmarkEmailSender()
    assert sender.api_token is not None
    assert sender.SMTP_SERVER == "smtp.postmarkapp.com"
    assert sender.SMTP_PORT == 587


@pytest.mark.flow
def test_full_email_flow(client: TacomailClient):
    # Get random email address
    username = client.get_random_username()
    domains = client.get_domains()
    domain = domains[0]
    test_email = f"{username}@{domain}"

    # Create session to receive emails (required in API v2)
    client.create_session(username, domain)

    # Verify inbox is empty initially
    initial_inbox = client.get_inbox(test_email)
    assert len(initial_inbox) == 0

    # Send test email
    sender = PostmarkEmailSender()
    test_subject = "Test Email Flow"
    test_body = "This is a test email for the full flow test."

    success = sender.send_email(
        to_email=test_email, subject=test_subject, body=test_body
    )
    assert success is True

    # Wait for email to arrive
    received_email = client.wait_for_email(test_email)
    assert received_email is not None

    # Verify email contents
    assert received_email.subject == test_subject
    assert received_email.body.text.strip() == test_body
    assert received_email.to.address == test_email


@pytest.mark.flow
def test_delete_functionality(client: TacomailClient):
    # Setup: Create email address and send two test emails
    username = client.get_random_username()
    domains = client.get_domains()
    domain = domains[0]
    test_email = f"{username}@{domain}"

    # Create session to receive emails (required in API v2)
    client.create_session(username, domain)

    sender = PostmarkEmailSender()

    # Send both emails
    sender.send_email(
        to_email=test_email, subject="Test Email 1", body="First test email"
    )

    sender.send_email(
        to_email=test_email, subject="Test Email 2", body="Second test email"
    )

    # Wait for both emails to arrive
    received_email = client.wait_for_email(test_email, expected_count=2)
    assert received_email is not None

    # Verify both emails are in inbox
    inbox = client.get_inbox(test_email)
    assert len(inbox) == 2

    # Delete first email
    first_email_id = inbox[0].id
    client.delete_email(test_email, first_email_id)

    # Verify only one email remains
    inbox = client.get_inbox(test_email)
    assert len(inbox) == 1
    assert inbox[0].id != first_email_id

    # Delete entire inbox
    client.delete_inbox(test_email)

    # Verify inbox is empty
    inbox = client.get_inbox(test_email)
    assert len(inbox) == 0


@pytest.mark.flow
def test_wait_for_email_timeout(client: TacomailClient):
    # Get random email address
    username = client.get_random_username()
    domains = client.get_domains()
    test_email = f"{username}@{domains[0]}"

    # Try to wait for an email that will never arrive (with short timeout)
    result = client.wait_for_email(test_email, timeout=5, interval=1)
    assert result is None


@pytest.mark.flow
def test_wait_for_email_filtered(client: TacomailClient):
    # Get random email address
    username = client.get_random_username()
    domains = client.get_domains()
    domain = domains[0]
    test_email = f"{username}@{domain}"

    # Create session to receive emails (required in API v2)
    client.create_session(username, domain)

    sender = PostmarkEmailSender()

    # Send two emails with different subjects
    sender.send_email(
        to_email=test_email, subject="First Subject", body="First test email"
    )

    sender.send_email(
        to_email=test_email, subject="Second Subject", body="Second test email"
    )

    # Wait for email with specific subject
    def filter_second_email(email: Email) -> bool:
        return email.subject == "Second Subject"

    received_email = client.wait_for_email_filtered(
        test_email, filter_fn=filter_second_email
    )

    assert received_email is not None
    assert received_email.subject == "Second Subject"
    assert received_email.body.text.strip() == "Second test email"

    # Test with a filter that won't match
    def filter_nonexistent(email: Email) -> bool:
        return email.subject == "Nonexistent Subject"

    result = client.wait_for_email_filtered(
        test_email, filter_fn=filter_nonexistent, timeout=5, interval=1
    )
    assert result is None



"""Tests for the Tacomail CLI output format functionality."""

import json
import pytest
from unittest.mock import patch, MagicMock
from datetime import datetime
from typer.testing import CliRunner

from tacomail.cli import app
from tacomail import Email, Session


runner = CliRunner()


@pytest.fixture
def mock_email():
    """Create a mock Email object for testing."""
    email = MagicMock(spec=Email)
    email.id = "test-email-id-123"
    email.subject = "Test Subject"
    email.date = datetime(2024, 1, 15, 10, 30, 0)
    email.from_ = MagicMock()
    email.from_.name = "Sender Name"
    email.from_.address = "sender@example.com"
    email.to = MagicMock()
    email.to.name = "Recipient Name"
    email.to.address = "recipient@example.com"
    email.body = MagicMock()
    email.body.text = "This is the email body."
    email.attachments = []
    return email


@pytest.fixture
def mock_session():
    """Create a mock Session object for testing."""
    session = MagicMock(spec=Session)
    session.username = "testuser"
    session.domain = "tacomail.de"
    session.expires = 1705320600000  # Some future timestamp in milliseconds
    return session


class TestCreatePlainOutput:
    """Tests for the 'create' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_create_plain_output(self, mock_get_client):
        """Test that create command outputs key=value format in plain mode."""
        mock_client = MagicMock()
        mock_client.get_random_address.return_value = "random123@tacomail.de"
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "create"])

        assert result.exit_code == 0
        assert "email=random123@tacomail.de" in result.stdout
        # Should not contain Rich formatting
        assert "✨" not in result.stdout
        assert "[bold" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_create_with_domain_plain_output(self, mock_get_client):
        """Test create command with domain option in plain mode."""
        mock_client = MagicMock()
        mock_client.get_random_username.return_value = "randomuser"
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "create", "-d", "tacomail.de"])

        assert result.exit_code == 0
        assert "email=randomuser@tacomail.de" in result.stdout


class TestCreateJsonOutput:
    """Tests for the 'create' command JSON output."""

    @patch("tacomail.cli.get_client")
    def test_create_json_output(self, mock_get_client):
        """Test that create command outputs JSON format."""
        mock_client = MagicMock()
        mock_client.get_random_address.return_value = "random123@tacomail.de"
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "create"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["email"] == "random123@tacomail.de"


class TestListDomainsPlainOutput:
    """Tests for the 'list-domains' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_list_domains_plain_output(self, mock_get_client):
        """Test that list-domains outputs one domain per line in plain mode."""
        mock_client = MagicMock()
        mock_client.get_domains.return_value = ["tacomail.de", "example.com", "test.org"]
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "list-domains"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert lines == ["tacomail.de", "example.com", "test.org"]
        # Should not contain Rich formatting
        assert "•" not in result.stdout
        assert "[bold" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_list_domains_json_output(self, mock_get_client):
        """Test that list-domains outputs JSON format."""
        mock_client = MagicMock()
        mock_client.get_domains.return_value = ["tacomail.de", "example.com", "test.org"]
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "list-domains"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data == ["tacomail.de", "example.com", "test.org"]


class TestCreateSessionPlainOutput:
    """Tests for the 'create-session' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_create_session_plain_output(self, mock_get_client, mock_session):
        """Test that create-session outputs key=value format in plain mode."""
        mock_client = MagicMock()
        mock_client.create_session.return_value = mock_session
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "create-session", "testuser@tacomail.de"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert "email=testuser@tacomail.de" in lines
        assert "username=testuser" in lines
        assert "domain=tacomail.de" in lines
        assert any(line.startswith("expires=") for line in lines)
        # Should not contain Rich formatting
        assert "🔐" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_create_session_json_output(self, mock_get_client, mock_session):
        """Test that create-session outputs JSON format."""
        mock_client = MagicMock()
        mock_client.create_session.return_value = mock_session
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "create-session", "testuser@tacomail.de"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["email"] == "testuser@tacomail.de"
        assert data["username"] == "testuser"
        assert data["domain"] == "tacomail.de"
        assert "expires" in data


class TestDeleteSessionPlainOutput:
    """Tests for the 'delete-session' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_delete_session_plain_output(self, mock_get_client):
        """Test that delete-session outputs deleted=email in plain mode."""
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "delete-session", "testuser@tacomail.de"])

        assert result.exit_code == 0
        assert result.stdout.strip() == "deleted=testuser@tacomail.de"
        # Should not contain Rich formatting
        assert "✓" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_delete_session_json_output(self, mock_get_client):
        """Test that delete-session outputs JSON format."""
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "delete-session", "testuser@tacomail.de"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["deleted"] == "testuser@tacomail.de"


class TestListInboxPlainOutput:
    """Tests for the 'list' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_list_inbox_plain_output(self, mock_get_client, mock_email):
        """Test that list outputs tab-separated values in plain mode."""
        mock_client = MagicMock()
        mock_client.get_inbox.return_value = [mock_email]
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "list", "testuser@tacomail.de"])

        assert result.exit_code == 0
        # Output should be tab-separated: id, from_address, subject, date
        line = result.stdout.strip()
        parts = line.split("\t")
        assert len(parts) == 4
        assert parts[0] == "test-email-id-123"
        assert parts[1] == "sender@example.com"
        assert parts[2] == "Test Subject"

    @patch("tacomail.cli.get_client")
    def test_list_inbox_empty_plain_output(self, mock_get_client):
        """Test that empty inbox produces no output in plain mode."""
        mock_client = MagicMock()
        mock_client.get_inbox.return_value = []
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "list", "testuser@tacomail.de"])

        assert result.exit_code == 0
        assert result.stdout.strip() == ""

    @patch("tacomail.cli.get_client")
    def test_list_inbox_json_output(self, mock_get_client, mock_email):
        """Test that list outputs JSON format."""
        mock_client = MagicMock()
        mock_client.get_inbox.return_value = [mock_email]
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "list", "testuser@tacomail.de"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert isinstance(data, list)
        assert len(data) == 1
        assert data[0]["id"] == "test-email-id-123"
        assert data[0]["from"] == "sender@example.com"


class TestGetEmailPlainOutput:
    """Tests for the 'get' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_get_email_plain_output(self, mock_get_client, mock_email):
        """Test that get outputs key=value format in plain mode."""
        mock_client = MagicMock()
        mock_client.get_email.return_value = mock_email
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "get", "testuser@tacomail.de", "email-id"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert "id=test-email-id-123" in lines
        assert "from=sender@example.com" in lines
        assert "from_name=Sender Name" in lines
        assert "to=recipient@example.com" in lines
        assert "to_name=Recipient Name" in lines
        assert "subject=Test Subject" in lines
        assert "body=This is the email body." in lines
        # Should not contain Rich formatting
        assert "📧" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_get_email_json_output(self, mock_get_client, mock_email):
        """Test that get outputs JSON format."""
        mock_client = MagicMock()
        mock_client.get_email.return_value = mock_email
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "get", "testuser@tacomail.de", "email-id"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["id"] == "test-email-id-123"
        assert data["from"] == "sender@example.com"
        assert data["subject"] == "Test Subject"
        assert data["body"] == "This is the email body."


class TestDeleteEmailPlainOutput:
    """Tests for the 'delete' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_delete_email_plain_output(self, mock_get_client):
        """Test that delete outputs deleted=id in plain mode."""
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "delete", "testuser@tacomail.de", "email-id-123"])

        assert result.exit_code == 0
        assert result.stdout.strip() == "deleted=email-id-123"

    @patch("tacomail.cli.get_client")
    def test_delete_email_json_output(self, mock_get_client):
        """Test that delete outputs JSON format."""
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "delete", "testuser@tacomail.de", "email-id-123"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["deleted"] == "email-id-123"


class TestClearInboxPlainOutput:
    """Tests for the 'clear' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_clear_inbox_plain_output(self, mock_get_client):
        """Test that clear outputs cleared=email in plain mode (no confirmation needed)."""
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "clear", "testuser@tacomail.de"])

        assert result.exit_code == 0
        assert result.stdout.strip() == "cleared=testuser@tacomail.de"

    @patch("tacomail.cli.get_client")
    def test_clear_inbox_json_output(self, mock_get_client):
        """Test that clear outputs JSON format."""
        mock_client = MagicMock()
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "clear", "testuser@tacomail.de"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["cleared"] == "testuser@tacomail.de"


class TestWaitPlainOutput:
    """Tests for the 'wait' command plain output."""

    @patch("tacomail.cli.get_client")
    def test_wait_plain_output_success(self, mock_get_client, mock_email):
        """Test that wait outputs key=value format in plain mode when email received."""
        mock_client = MagicMock()
        mock_client.wait_for_email.return_value = mock_email
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "wait", "testuser@tacomail.de", "-t", "1"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert "id=test-email-id-123" in lines
        assert "from=sender@example.com" in lines
        assert "subject=Test Subject" in lines
        # Should not have the waiting message
        assert "Waiting for email" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_wait_plain_output_timeout(self, mock_get_client):
        """Test that wait outputs timeout=True in plain mode when no email."""
        mock_client = MagicMock()
        mock_client.wait_for_email.return_value = None
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "wait", "testuser@tacomail.de", "-t", "1"])

        assert result.exit_code == 1
        assert "timeout=True" in result.stdout

    @patch("tacomail.cli.get_client")
    def test_wait_plain_output_with_body(self, mock_get_client, mock_email):
        """Test that wait with --print-body includes body in plain mode."""
        mock_client = MagicMock()
        mock_client.wait_for_email.return_value = mock_email
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "wait", "testuser@tacomail.de", "-t", "1", "--print-body"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert "body=This is the email body." in lines

    @patch("tacomail.cli.get_client")
    def test_wait_json_output_success(self, mock_get_client, mock_email):
        """Test that wait outputs JSON format when email received."""
        mock_client = MagicMock()
        mock_client.wait_for_email.return_value = mock_email
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "wait", "testuser@tacomail.de", "-t", "1"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["id"] == "test-email-id-123"
        assert data["from"] == "sender@example.com"
        assert data["subject"] == "Test Subject"

    @patch("tacomail.cli.get_client")
    def test_wait_json_output_timeout(self, mock_get_client):
        """Test that wait outputs JSON format on timeout."""
        mock_client = MagicMock()
        mock_client.wait_for_email.return_value = None
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "wait", "testuser@tacomail.de", "-t", "1"])

        assert result.exit_code == 1
        data = json.loads(result.stdout.strip())
        assert data["timeout"] is True


class TestCreateWithSessionPlainOutput:
    """Tests for the 'create-with-session' and 'new' commands plain output."""

    @patch("tacomail.cli.get_client")
    def test_create_with_session_plain_output(self, mock_get_client, mock_session):
        """Test that create-with-session outputs key=value format in plain mode."""
        mock_client = MagicMock()
        mock_client.get_random_address.return_value = "random123@tacomail.de"
        mock_client.create_session.return_value = mock_session
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "create-with-session"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert "email=random123@tacomail.de" in lines
        assert "username=testuser" in lines
        assert "domain=tacomail.de" in lines
        assert any(line.startswith("expires=") for line in lines)
        # Should not contain Rich formatting or next steps
        assert "✨" not in result.stdout
        assert "Next steps" not in result.stdout

    @patch("tacomail.cli.get_client")
    def test_new_command_plain_output(self, mock_get_client, mock_session):
        """Test that 'new' command (alias) also works with plain output."""
        mock_client = MagicMock()
        mock_client.get_random_address.return_value = "random123@tacomail.de"
        mock_client.create_session.return_value = mock_session
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "plain", "new"])

        assert result.exit_code == 0
        lines = result.stdout.strip().split("\n")
        assert "email=random123@tacomail.de" in lines

    @patch("tacomail.cli.get_client")
    def test_create_with_session_json_output(self, mock_get_client, mock_session):
        """Test that create-with-session outputs JSON format."""
        mock_client = MagicMock()
        mock_client.get_random_address.return_value = "random123@tacomail.de"
        mock_client.create_session.return_value = mock_session
        mock_get_client.return_value = mock_client

        result = runner.invoke(app, ["--output", "json", "create-with-session"])

        assert result.exit_code == 0
        data = json.loads(result.stdout.strip())
        assert data["email"] == "random123@tacomail.de"
        assert data["username"] == "testuser"
        assert data["domain"] == "tacomail.de"
        assert "expires" in data


class TestOutputOptionHelp:
    """Test that the --output option is properly documented."""

    def test_output_option_in_help(self):
        """Test that --output option appears in CLI help."""
        result = runner.invoke(app, ["--help"])

        assert result.exit_code == 0
        # Output may contain ANSI escape codes, so search for keywords
        assert "output" in result.stdout.lower()
        assert "json" in result.stdout.lower()
        assert "plain" in result.stdout.lower()
        assert "rich" in result.stdout.lower()



from .tacomail import (
    TacomailClient,
    AsyncTacomailClient,
    Email,
    EmailAddress,
    EmailBody,
    Attachment,
    Session,
)

__all__ = [
    "TacomailClient",
    "Email",
    "EmailAddress",
    "EmailBody",
    "Attachment",
    "AsyncTacomailClient",
    "Session",
]


import os
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from dotenv import load_dotenv


class PostmarkEmailSender:
    SMTP_SERVER = "smtp.postmarkapp.com"
    SMTP_PORT = 587

    def __init__(self):
        load_dotenv()
        api_token = os.getenv("POSTMARK_API_TOKEN")
        sender_email = os.getenv("SENDER_EMAIL")
        if api_token is None:
            raise ValueError("POSTMARK_API_TOKEN not found in environment variables")
        self.api_token = api_token
        if sender_email is None:
            raise ValueError("SENDER_EMAIL not found in environment variables")
        self.sender_email = sender_email

    def send_email(self, to_email: str, subject: str, body: str):
        """Send an email using Postmark SMTP server."""
        msg = MIMEMultipart()
        msg["From"] = self.sender_email
        msg["To"] = to_email
        msg["Subject"] = subject

        msg.attach(MIMEText(body, "plain"))

        try:
            with smtplib.SMTP(self.SMTP_SERVER, self.SMTP_PORT) as server:
                server.starttls()  # Enable TLS
                server.login(
                    self.api_token, self.api_token
                )  # Using API token as both username and password
                server.send_message(msg)
                return True
        except Exception as e:
            print(f"Failed to send email: {str(e)}")
            return False



from typing import List, Optional, Union, Callable
import httpx
from dataclasses import dataclass
from datetime import datetime
import time
import asyncio
import random


@dataclass
class EmailAddress:
    address: str
    name: str


@dataclass
class EmailBody:
    text: str
    html: str


@dataclass
class Attachment:
    id: str
    fileName: str
    present: bool


@dataclass
class Session:
    expires: int
    username: str
    domain: str


@dataclass
class Email:
    id: str
    from_: EmailAddress
    to: EmailAddress
    subject: str
    date: datetime
    body: EmailBody
    headers: dict
    attachments: List[Attachment]

    @classmethod
    def from_dict(cls, data: dict) -> "Email":
        return cls(
            id=data["id"],
            from_=EmailAddress(**data["from"]),
            to=EmailAddress(**data["to"]),
            subject=data["subject"],
            date=datetime.fromisoformat(data["date"].replace("Z", "+00:00")),
            body=EmailBody(**data["body"]),
            headers=data["headers"],
            attachments=[Attachment(**att) for att in data["attachments"]],
        )


class TacomailClient:
    """A synchronous client for interacting with the TacoMail API.

    This client provides methods to interact with temporary email inboxes, including
    retrieving emails, managing attachments, and waiting for new messages.

    Args:
        base_url (str): The base URL of the TacoMail API. Defaults to "https://tacomail.de".
        client (Optional[httpx.Client]): An optional pre-configured httpx Client instance.
            If not provided, a new client will be created.

    Examples:
        >>> client = TacomailClient()
        >>> # Get a random username and available domains
        >>> username = client.get_random_username()
        >>> domains = client.get_domains()
        >>> # Create an email address and check for emails
        >>> address = f"{username}@{domains[0]}"
        >>> emails = client.get_inbox(address)
    """

    def __init__(
        self,
        base_url: str = "https://tacomail.de",
        client: Optional[httpx.Client] = None,
    ):
        self.base_url = base_url.rstrip("/")
        self.client = client if client is not None else httpx.Client()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()

    def get_contact_email(self) -> str:
        """Get the contact email address of the TacoMail instance."""
        response = self.client.get(f"{self.base_url}/api/v2/contactEmail")
        response.raise_for_status()
        return response.json()["email"]

    def get_random_username(self) -> str:
        """Get a random username."""
        response = self.client.get(f"{self.base_url}/api/v2/randomUsername")
        response.raise_for_status()
        return response.json()["username"]

    def get_domains(self) -> List[str]:
        """Get all available domains."""
        response = self.client.get(f"{self.base_url}/api/v2/domains")
        response.raise_for_status()
        return response.json()

    def create_session(self, username: str, domain: str) -> Session:
        """Create an inbox session for receiving emails.
        
        Only incoming mails with an associated inbox session are saved.
        A session is valid for a configured time and can be renewed by
        calling this method again with the same credentials.
        
        Args:
            username: The username part of the email address
            domain: The domain part of the email address
            
        Returns:
            Session object containing expiration timestamp and credentials
        """
        response = self.client.post(
            f"{self.base_url}/api/v2/session",
            json={"username": username, "domain": domain}
        )
        response.raise_for_status()
        return Session(**response.json())

    def delete_session(self, username: str, domain: str) -> None:
        """Delete an inbox session.
        
        This will cause any incoming mails associated with the session's
        address to be rejected. It does not delete already saved mails.
        
        Args:
            username: The username part of the email address
            domain: The domain part of the email address
        """
        response = self.client.request(
            "DELETE",
            f"{self.base_url}/api/v2/session",
            json={"username": username, "domain": domain}
        )
        response.raise_for_status()

    def get_random_address(self) -> str:
        """Get a random email address."""
        username = self.get_random_username()
        domains = self.get_domains()
        random_domain = random.choice(domains)
        return f"{username}@{random_domain}"

    def get_inbox(self, address: str, limit: Optional[int] = None) -> List[Email]:
        """Get the last 10 emails (or specified limit) from an inbox.
        
        Note: The maximum number of emails that can be retrieved is 10, regardless of the limit parameter."""
        params = {"limit": limit} if limit is not None else None
        response = self.client.get(
            f"{self.base_url}/api/v2/mail/{address}", params=params
        )
        response.raise_for_status()
        return [Email.from_dict(email) for email in response.json()]

    def get_email(self, address: str, mail_id: str) -> Email:
        """Get a single email by its ID."""
        response = self.client.get(f"{self.base_url}/api/v2/mail/{address}/{mail_id}")
        response.raise_for_status()
        return Email.from_dict(response.json())

    def get_attachments(self, address: str, mail_id: str) -> List[Attachment]:
        """Get all attachments of an email."""
        response = self.client.get(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}/attachments"
        )
        response.raise_for_status()
        return [Attachment(**att) for att in response.json()]

    def download_attachment(
        self, address: str, mail_id: str, attachment_id: str
    ) -> bytes:
        """Download a single attachment."""
        response = self.client.get(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}/attachments/{attachment_id}"
        )
        response.raise_for_status()
        return response.content

    def delete_email(self, address: str, mail_id: str) -> None:
        """Delete a single email."""
        response = self.client.delete(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}"
        )
        response.raise_for_status()

    def delete_inbox(self, address: str) -> None:
        """Delete all emails from an inbox."""
        response = self.client.delete(f"{self.base_url}/api/v2/mail/{address}")
        response.raise_for_status()

    def wait_for_email(
        self,
        address: str,
        timeout: int = 30,
        interval: int = 2,
        expected_count: int = 1,
    ) -> Union[Email, None]:
        """
        Wait for an email to arrive in the inbox.

        Args:
            address: Email address to check
            timeout: Maximum time to wait in seconds (default: 30)
            interval: Time between checks in seconds (default: 2)
            expected_count: Expected number of emails in inbox (default: 1)

        Returns:
            The latest Email object if found, None if timeout is reached
        """
        attempts = timeout // interval

        for _ in range(attempts):
            inbox = self.get_inbox(address)
            if len(inbox) >= expected_count:
                return inbox[0]  # Return the most recent email
            time.sleep(interval)

        return None

    def wait_for_email_filtered(
        self,
        address: str,
        filter_fn: Callable[[Email], bool],
        timeout: int = 30,
        interval: int = 2,
    ) -> Union[Email, None]:
        """
        Wait for an email that matches the given filter function.

        Args:
            address: Email address to check
            filter_fn: Function that takes an Email object and returns bool
            timeout: Maximum time to wait in seconds (default: 30)
            interval: Time between checks in seconds (default: 2)

        Returns:
            The first Email object that matches the filter, None if timeout is reached
        """
        attempts = timeout // interval

        for _ in range(attempts):
            inbox = self.get_inbox(address)
            for email in inbox:
                if filter_fn(email):
                    return email
            time.sleep(interval)

        return None


class AsyncTacomailClient:
    def __init__(
        self,
        base_url: str = "https://tacomail.de",
        client: Optional[httpx.AsyncClient] = None,
    ):
        self.base_url = base_url.rstrip("/")
        self.client = client if client is not None else httpx.AsyncClient()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.client.aclose()

    async def get_contact_email(self) -> str:
        """Get the contact email address of the TacoMail instance."""
        response = await self.client.get(f"{self.base_url}/api/v2/contactEmail")
        response.raise_for_status()
        return response.json()["email"]

    async def get_random_username(self) -> str:
        """Get a random username."""
        response = await self.client.get(f"{self.base_url}/api/v2/randomUsername")
        response.raise_for_status()
        return response.json()["username"]

    async def get_domains(self) -> List[str]:
        """Get all available domains."""
        response = await self.client.get(f"{self.base_url}/api/v2/domains")
        response.raise_for_status()
        return response.json()

    async def create_session(self, username: str, domain: str) -> Session:
        """Create an inbox session for receiving emails.
        
        Only incoming mails with an associated inbox session are saved.
        A session is valid for a configured time and can be renewed by
        calling this method again with the same credentials.
        
        Args:
            username: The username part of the email address
            domain: The domain part of the email address
            
        Returns:
            Session object containing expiration timestamp and credentials
        """
        response = await self.client.post(
            f"{self.base_url}/api/v2/session",
            json={"username": username, "domain": domain}
        )
        response.raise_for_status()
        return Session(**response.json())

    async def delete_session(self, username: str, domain: str) -> None:
        """Delete an inbox session.
        
        This will cause any incoming mails associated with the session's
        address to be rejected. It does not delete already saved mails.
        
        Args:
            username: The username part of the email address
            domain: The domain part of the email address
        """
        response = await self.client.request(
            "DELETE",
            f"{self.base_url}/api/v2/session",
            json={"username": username, "domain": domain}
        )
        response.raise_for_status()

    async def get_random_address(self) -> str:
        """Get a random email address."""
        username, domains = await asyncio.gather(
            self.get_random_username(), self.get_domains()
        )
        random_domain = random.choice(domains)
        return f"{username}@{random_domain}"

    async def get_inbox(self, address: str, limit: Optional[int] = None) -> List[Email]:
        """Get the last 10 emails (or specified limit) from an inbox.
        
        Note: The maximum number of emails that can be retrieved is 10, regardless of the limit parameter."""
        params = {"limit": limit} if limit is not None else None
        response = await self.client.get(
            f"{self.base_url}/api/v2/mail/{address}", params=params
        )
        response.raise_for_status()
        return [Email.from_dict(email) for email in response.json()]

    async def get_email(self, address: str, mail_id: str) -> Email:
        """Get a single email by its ID."""
        response = await self.client.get(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}"
        )
        response.raise_for_status()
        return Email.from_dict(response.json())

    async def get_attachments(self, address: str, mail_id: str) -> List[Attachment]:
        """Get all attachments of an email."""
        response = await self.client.get(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}/attachments"
        )
        response.raise_for_status()
        return [Attachment(**att) for att in response.json()]

    async def download_attachment(
        self, address: str, mail_id: str, attachment_id: str
    ) -> bytes:
        """Download a single attachment."""
        response = await self.client.get(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}/attachments/{attachment_id}"
        )
        response.raise_for_status()
        return response.content

    async def delete_email(self, address: str, mail_id: str) -> None:
        """Delete a single email."""
        response = await self.client.delete(
            f"{self.base_url}/api/v2/mail/{address}/{mail_id}"
        )
        response.raise_for_status()

    async def delete_inbox(self, address: str) -> None:
        """Delete all emails from an inbox."""
        response = await self.client.delete(f"{self.base_url}/api/v2/mail/{address}")
        response.raise_for_status()

    async def wait_for_email(
        self,
        address: str,
        timeout: int = 30,
        interval: float = 2,
        expected_count: int = 1,
    ) -> Union[Email, None]:
        """
        Wait for an email to arrive in the inbox.

        Args:
            address: Email address to check
            timeout: Maximum time to wait in seconds (default: 30)
            interval: Time between checks in seconds (default: 2)
            expected_count: Expected number of emails in inbox (default: 1)

        Returns:
            The latest Email object if found, None if timeout is reached
        """
        end_time = time.time() + timeout

        while True:
            inbox = await self.get_inbox(address)
            if len(inbox) >= expected_count:
                return inbox[0]  # Return the most recent email

            # Calculate remaining time to avoid oversleeping
            remaining = end_time - time.time()
            if remaining <= 0:
                break

            # Sleep for the interval or remaining time, whichever is shorter
            await asyncio.sleep(min(interval, remaining))

        return None

    async def wait_for_email_filtered(
        self,
        address: str,
        filter_fn: Callable[[Email], bool],
        timeout: int = 30,
        interval: float = 2,
    ) -> Union[Email, None]:
        """
        Wait for an email that matches the given filter function.

        Args:
            address: Email address to check
            filter_fn: Function that takes an Email object and returns bool
            timeout: Maximum time to wait in seconds (default: 30)
            interval: Time between checks in seconds (default: 2)

        Returns:
            The first Email object that matches the filter, None if timeout is reached
        """
        end_time = time.time() + timeout

        while True:
            inbox = await self.get_inbox(address)
            for email in inbox:
                if filter_fn(email):
                    return email

            # Calculate remaining time to avoid oversleeping
            remaining = end_time - time.time()
            if remaining <= 0:
                break

            # Sleep for the interval or remaining time, whichever is shorter
            await asyncio.sleep(min(interval, remaining))

        return None



import pytest
from tacomail import AsyncTacomailClient, Email, Session
from email_sender import PostmarkEmailSender
import time
from typing import AsyncGenerator

pytest_plugins = ("pytest_asyncio",)


@pytest.fixture
async def client_generator():
    async with AsyncTacomailClient("https://tacomail.de") as client:
        yield client


@pytest.mark.asyncio
async def test_create_session_async(
    client_generator: AsyncGenerator[AsyncTacomailClient, None],
):
    client: AsyncTacomailClient = await anext(client_generator)

    username = await client.get_random_username()
    domains = await client.get_domains()
    domain = domains[0]

    session = await client.create_session(username, domain)

    assert isinstance(session, Session)
    assert session.username == username
    assert session.domain == domain
    assert isinstance(session.expires, int)
    assert session.expires > 0


@pytest.mark.asyncio
async def test_delete_session_async(
    client_generator: AsyncGenerator[AsyncTacomailClient, None],
):
    client: AsyncTacomailClient = await anext(client_generator)

    username = await client.get_random_username()
    domains = await client.get_domains()
    domain = domains[0]

    # Create a session first
    await client.create_session(username, domain)

    # Delete the session (should not raise an exception)
    await client.delete_session(username, domain)


@pytest.mark.flow
@pytest.mark.asyncio
async def test_wait_for_email_async(
    client_generator: AsyncGenerator[AsyncTacomailClient, None],
):
    client: AsyncTacomailClient = await anext(client_generator)

    # Get random email address
    username = await client.get_random_username()
    domains = await client.get_domains()
    domain = domains[0]
    test_email = f"{username}@{domain}"

    # Create session to receive emails (required in API v2)
    await client.create_session(username, domain)

    # Send test email (using sync sender)
    sender = PostmarkEmailSender()
    test_subject = "Test Async Email"
    test_body = "This is a test email for the async flow test."

    success = sender.send_email(
        to_email=test_email, subject=test_subject, body=test_body
    )
    assert success is True

    # Wait for email asynchronously
    received_email = await client.wait_for_email(test_email)
    assert received_email is not None
    assert received_email.subject == test_subject
    assert received_email.body.text.strip() == test_body


@pytest.mark.flow
@pytest.mark.asyncio
async def test_wait_for_email_filtered_async(
    client_generator: AsyncGenerator[AsyncTacomailClient, None],
):
    client: AsyncTacomailClient = await anext(client_generator)

    # Get random email address
    username = await client.get_random_username()
    domains = await client.get_domains()
    domain = domains[0]
    test_email = f"{username}@{domain}"

    # Create session to receive emails (required in API v2)
    await client.create_session(username, domain)

    sender = PostmarkEmailSender()

    # Send two emails with different subjects
    sender.send_email(
        to_email=test_email,
        subject="First Async Subject",
        body="First async test email",
    )

    sender.send_email(
        to_email=test_email,
        subject="Second Async Subject",
        body="Second async test email",
    )

    # Wait for email with specific subject
    def filter_second_email(email: Email) -> bool:
        return email.subject == "Second Async Subject"

    received_email = await client.wait_for_email_filtered(
        test_email, filter_fn=filter_second_email
    )

    assert received_email is not None
    assert received_email.subject == "Second Async Subject"
    assert received_email.body.text.strip() == "Second async test email"


@pytest.mark.asyncio
async def test_wait_for_email_timeout_async(
    client_generator: AsyncGenerator[AsyncTacomailClient, None],
):
    client: AsyncTacomailClient = await anext(client_generator)

    # Get random email address
    username = await client.get_random_username()
    domains = await client.get_domains()
    test_email = f"{username}@{domains[0]}"

    # Try to wait for an email that will never arrive
    result = await client.wait_for_email(test_email, timeout=5, interval=1)
    assert result is None


@pytest.mark.asyncio
async def test_wait_for_email_timing_async(
    client_generator: AsyncGenerator[AsyncTacomailClient, None],
):
    client: AsyncTacomailClient = await anext(client_generator)

    # Get random email address
    username = await client.get_random_username()
    domains = await client.get_domains()
    test_email = f"{username}@{domains[0]}"

    # Record start time
    start_time = time.time()

    # Try to wait for an email that will never arrive
    result = await client.wait_for_email(test_email, timeout=3, interval=0.5)

    # Calculate elapsed time
    elapsed = time.time() - start_time

    assert result is None
    # Allow for small timing variations but ensure we're close to the timeout
    assert 2.9 <= elapsed <= 4.0, f"Expected timeout of 3 seconds, got {elapsed}"
